import os
import json


def get_parent_dir(path, level=1):
    """ returns the absolute path to parent directory
    
    level -- How many levels up the tree you wish to go

    """
    parent = path

    for _ in range(level):
        parent = os.path.dirname(parent)

    return parent

def load_all_clients(path:str):
    """ returns a json of all clients from a provided path
    """
    clients = [client.replace(".json", "") for client in os.listdir(path) if ".json" in client]

    clients = {"clients":clients}

    return json.dumps(clients, indent=1)




from iqmotion.iq_devices.iq_module import IqModule
from iqmotion.communication.serial_communicator import SerialCommunicator
from iqmotion.custom_errors import CommunicationError, IqModuleError
from iqmotion.iq_devices.utils import get_parent_dir, load_all_clients

import os
import sys
import glob
import serial

SPEED   = '1'
SERVO   = '2'
STEPDIR = '4'

class RdModule(IqModule):
    """ Creates R&D Object with every publicly available json

    The R&D Module does a few things different than other modules:

        1. Creates dynamic json that contains all the public facing clients
        2. Automatically connects to serial com port
            Will throw an error if more than one serial port exists and one wasn't provided
    
    Optional Arguments:
        com {Communicator} -- The communicator object to interface with the IqModule
        port {str} -- will connect to this port if more than one port is detected
    
    Keyword Arguments:
        module_idn {int} -- The idn of the module (default: {0})
        extra_clients {list} -- list of file paths to extra clients you want to load in the module (default: {None})
    """

    def __init__(
        self,
        port: str=None,
        baudrate: int=115200,
        module_idn: int = 0,
        style: str = None,      
        clients_path: str = None
    ):

        self._DEFAULT_VELOCITY_CLIENT_ENTRY = "ctrl_velocity"
        self._DEFAULT_VOLTS_CLIENT_ENTRY = "ctrl_volts"
        self._MODULE_FILE_NAME = "rd.json"


        # Dynamically Load all clients into one module file
        path = os.path.join(get_parent_dir(__file__, 2), "clients/client_files")
        rd = os.path.join(os.path.dirname(__file__), "module_files/rd.json")
        clients = load_all_clients(path)
        
        # Load them into a RD Module file
        with open(rd, 'w') as f:
            f.write(clients)

        # Pull all the available ports into a list
        ports_avail = self._find_serial_ports()

        # Let user decide which port they want
        if port is not None:
            com = SerialCommunicator(port, baudrate)
        # Otherwise allow RdModule auto-connect if it's only 1 port open
        elif len(ports_avail) == 1:
            port = ports_avail[0]
            com = SerialCommunicator(port, baudrate)
        # If too many ports are available, you should pass one in
        elif len(ports_avail) > 1:
            err_msg = "Too many ports available to choose from, "
            err_msg += "Please choose one of the following:\n\n"
            err_msg += f"{ports_avail}\n"
            err_msg += "ex: motor = iq.RdModule(port='COM1', baudrate=115200)"
            raise CommunicationError(err_msg)    
        # No ports available
        else:
            err_msg = "No available ports detected"
            raise CommunicationError(err_msg)
        

        super().__init__(com, module_idn, clients_path)

        # Try to automagically figure out the firmware style
        firmware_version = self.get("system_control", "firmware_version")
        firmware_style = str(0xFFFF & (firmware_version >> 20))
        if (style=="speed" or firmware_style == SPEED):
            self._DEFAULT_CONTROL_CLIENT = "propeller_motor_control"
        elif(style=="position" or firmware_style == SERVO or firmware_style == STEPDIR):
            self._DEFAULT_CONTROL_CLIENT = "multi_turn_angle_control"
        else:
            err_msg = "Please select the motor firmware style:\n\n"
            err_msg += "'speed' or 'position'"
            raise IqModuleError(err_msg)


    def _find_serial_ports(self):
        """ Finds a list of open serial ports

        Raises:
            EnvironmentError: Unsupported platform

        Returns:
            list: List of str of open serial ports
        """

        if sys.platform.startswith("win"):
            ports = ["COM%s" % (i + 1) for i in range(256)]
        elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
            # this excludes your current terminal "/dev/tty"
            # this excludes everything but the /dev/ttyUSB
            ports = glob.glob("/dev/tty[U-Z]*")
        elif sys.platform.startswith("darwin"):
            ports = glob.glob("/dev/tty.*")
        else:
            raise EnvironmentError("Unsupported platform")

        working_ports = []
        for port in ports:
            try:
                s = serial.Serial(port)
                s.close()
                working_ports.append(port)
            except (OSError, serial.SerialException):
                pass

        return working_ports


from dataclasses import dataclass
import numpy as np

from iqmotion.message_making.message_maker import MessageMaker
from iqmotion.client_entries.dictionary_client_entry import DictionaryClientEntry
from iqmotion.client_entries.dictionary_client_entry import AccessType
from iqmotion.client_entries.dictionary_client_entry import DictionaryClientEntryData
from iqmotion.custom_errors import MessageMakerError


@dataclass
class DictionaryMessageMakerData:
    type_idn: bytes
    param_idn: bytes
    module_idn: bytes
    access_type: AccessType
    format: str
    value: any

    def __init__(
        self,
        client_entry_data: DictionaryClientEntryData,
        module_idn: bytes,
        access_type: AccessType,
        value,
    ):

        self.type_idn = client_entry_data.type_idn
        self.param_idn = client_entry_data.param_idn
        self.module_idn = module_idn
        self.access_type = access_type
        self.format = client_entry_data.format
        self.values = value


class DictionaryMessageMaker(MessageMaker):
    """ DictionaryMessageMaker is an implementation of MessageMaker
    This class defines how to make a message from a dictionary client entry
    """

    def __init__(
        self,
        client_entry: DictionaryClientEntry,
        module_idn: bytes,
        access_type: AccessType,
        values,
    ):
        """ Returns a DictonaryMessageMaker object to create dictionary client entry messages

        Args:
            client_entry (DictionaryClientEntry): client entry used to make the message
            module_idn (bytes): id of the module
            access_type (AccessType): Access type of the message
            values (int/list): values to be put in the message
        """

        self._data = DictionaryMessageMakerData(
            client_entry.data, module_idn, access_type, values
        )

        self._value_types = {
            "": None,
            "B": np.uint8,
            "b": np.int8,
            "f": np.float32,
            "H": np.uint16,
            "h": np.int16,
            "I": np.uint32,
            "i": np.int32,
        }

    def make(self):
        """ Makes the message

        Returns:
            bytearray: the message made

        Raises:
            MessageMakerError: if values too long for client entry
        """
        message = bytearray([self._data.type_idn])
        payload = self._make_payload()
        message.extend(payload)

        return message

    def _make_payload(self):
        param_idn = self._data.param_idn
        module_idn = self._data.module_idn
        access_type = self._data.access_type
        data_format = self._data.format
        values = self._data.values

        access = access_type.value + (module_idn * 4)
        payload = bytearray([param_idn, access])
        values_bytes = self._format_values(list(data_format), values)
        payload.extend(values_bytes)

        return payload

    def _format_values(self, format_list, values):
        if isinstance(values, list) and len(values) > 1:
            formated_values = self._format_values_list(format_list, values)
        elif values is None:
            formated_values = []
        else:
            formated_values = self._format_single_value(format_list[0], values)

        return bytearray(formated_values)

    def _format_values_list(self, format_list, values_list):
        values_bytes_array = bytearray([])
        for i, values in enumerate(values_list):
            data_format = self._parse_format_list(format_list, i)

            formated_values = self._value_types[data_format](values)
            values_bytes = formated_values.tobytes()

            values_bytes_array.extend(values_bytes)

        return values_bytes_array

    def _parse_format_list(self, format_list, current_index):
        try:
            if current_index >= len(format_list) - 1 and format_list[-1] == "*":
                data_format = format_list[-2]
            else:
                data_format = format_list[current_index]
        except IndexError:
            raise MessageMakerError("values too long for client entry")

        return data_format

    def _format_single_value(self, data_format, value):
        formated_value = self._value_types[data_format](value)
        value_byte = formated_value.tobytes()

        return value_byte



from abc import ABC, abstractmethod


class Communicator(ABC):
    """ Communicator is an interface class that enables different type of communication protocole
    implementations while keeping the same message structure for every device.
    It is the implementation's job to "package" the message into its own package as well as 
    extract messages from its own packaging scheme.
    """

    @abstractmethod
    def send_message(self, message: bytearray):
        """ this property is too abstract to understand. """

    @abstractmethod
    def add_to_out_queue(self, out_bytes: bytearray):
        """ this property is too abstract to understand. """

    @abstractmethod
    def send_now(self):
        """ this property is too abstract to understand. """

    @abstractmethod
    def flush_input_buffer(self):
        """ this property is too abstract to understand. """

    @abstractmethod
    def read_bytes(self):
        """ this property is too abstract to understand. """

    @property
    @abstractmethod
    def bytes_left_in_queue(self):
        """ this property is too abstract to understand. """

    @abstractmethod
    def extract_message(self):
        """ this property is too abstract to understand. """



from iqmotion.iq_devices.iq_module import IqModule
from iqmotion.communication.communicator import Communicator
from iqmotion.custom_errors import IqModuleError


# This is a IQ Module Wrapper for the vertiq 4006
# This Wrapper takes in firmware argument and creates an veritq4006 API Object
class Vertiq4006(IqModule):
    """ Creates vertiq4006 object with default speed firmware

    Arguments:
        com {Communicator} -- The communicator object to interface with the IqModule
        firmware {str} -- Which firmware is the motor using (default: {"speed"})

    Keyword Arguments:
        module_idn {int} -- The idn of the module (default: {0})
        extra_clients {list} -- list of file paths to extra clients you want to load in the module (default: {None})
    """

    def __init__(
            self,
            com: Communicator,
            module_idn: int = 0,
            firmware: str = "speed",  # Default Firmware
            clients_path: str = None
    ):
        if firmware.lower() == "speed":
            self._DEFAULT_CONTROL_CLIENT = "propeller_motor_control"
            self._DEFAULT_VELOCITY_CLIENT_ENTRY = "ctrl_velocity"
            self._DEFAULT_VOLTS_CLIENT_ENTRY = "ctrl_volts"
            self._MODULE_FILE_NAME = "speed4006.json"
        else:
            raise IqModuleError("'" + str(firmware) + "' firmware is not supported")

        super().__init__(com, module_idn, clients_path)



from iqmotion.iq_devices.iq_module import IqModule


class SpeedModule(IqModule):
    """ Creates SpeedModule object
    
    Arguments:
        com {Communicator} -- The communicator object to interface with the IqModule
    
    Keyword Arguments:
        module_idn {int} -- The idn of the module (default: {0})
        extra_clients {list} -- list of file paths to extra clients you want to load in the module (default: {None})
    """

    _DEFAULT_CONTROL_CLIENT = "propeller_motor_control"
    _DEFAULT_VELOCITY_CLIENT_ENTRY = "ctrl_velocity"
    _DEFAULT_VOLTS_CLIENT_ENTRY = "ctrl_volts"

    _MODULE_FILE_NAME = "speed.json"




from iqmotion.communication.packet_state import PacketState


class PacketParser:
    """ PacketParser is the "context" class of a State Pattern for Packet States.
    It defines a single interface to parse a message from a packet using the correct State implementations
    """

    def __init__(self, state: PacketState):
        """ Creates a PacketParser object to interface with PacketState's implementation (depending on the com/packet queues)

        Arguments:
            state (PacketState): the PacketState implementation you want to use to parse your packet
        """
        self._current_state = state
        self._parsing_successful = False

    @property
    def succesful(self):
        return self._current_state.is_succesful

    @property
    def message(self):
        """ The message parsed by the PacketStates
        """
        return self._current_state.message

    def parse(self):
        """ Parse packet from initialized PacketState until PacketState is done
        """
        while not self._current_state.is_done:
            self._find_next_state()

    def _find_next_state(self):
        self._current_state.parse()
        self._current_state = self._current_state.find_next_state()



from abc import ABC, abstractmethod


class MessageMaker(ABC):
    """ MessageMaker is the interface class used by the implementations of the bridge pattern.
    Each Implementations define how to make a message
    """

    @abstractmethod
    def make(self):
        """ this property is too abstract to understand. """



from iqmotion.iq_devices.base_iq_module import BaseIqModule
from iqmotion.iq_devices.custom_iq_module import CustomIqModule
from iqmotion.iq_devices.rd_iq_module import RdModule

# Soon to be deprecated: but the original way to access motor firmware API
from iqmotion.iq_devices.speed_module import SpeedModule
from iqmotion.iq_devices.servo_module import ServoModule
from iqmotion.iq_devices.step_dir_module import StepDirModule
from iqmotion.iq_devices.pulsing_module import PulsingModule

# Make Motor Modules front facing visible
from iqmotion.iq_devices.vertiq8108_module import Vertiq8108
from iqmotion.iq_devices.vertiq4006_module import Vertiq4006
from iqmotion.iq_devices.vertiq2306_module import Vertiq2306
from iqmotion.iq_devices.fortiq_module import Fortiq

from iqmotion.communication.communicator import Communicator
from iqmotion.communication.serial_communicator import SerialCommunicator



from iqmotion.iq_devices.iq_module import IqModule
from iqmotion.communication.communicator import Communicator
from iqmotion.custom_errors import IqModuleError


class Vertiq2306(IqModule):
    def __init__(
        self,
        com: Communicator,
        module_idn: int = 0,
        firmware: str = "speed",  # Default Firmware
        clients_path: str = None,
    ):
        """Creates Vertiq2306 object with default speed firmware.

        :param com: The communicator object used to interface with the IqModule
        :type com: Communicator
        :param module_idn: The idn of the module (default: 0)
        :type module_idn: int
        :param firmware: The firmware type used to determine which clients are available for the module to use
            Options include:
                * speed
                * stepdir
                * servo
                * pulsing
        :type firmware: str
        :param clients_path: Optional parameter to specify a directory containing extra clients for the module to use (default: {None})
        :type clients_path: str
        """

        # Point to the correct JSON File depending on the firmware
        if firmware.lower() == "speed":
            self._DEFAULT_CONTROL_CLIENT = "propeller_motor_control"
            self._DEFAULT_VELOCITY_CLIENT_ENTRY = "ctrl_velocity"
            self._DEFAULT_VOLTS_CLIENT_ENTRY = "ctrl_volts"

            self._MODULE_FILE_NAME = "speed.json"

        elif firmware.lower() == "stepdir":
            self._DEFAULT_CONTROL_CLIENT = "multi_turn_angle_control"
            self._DEFAULT_VELOCITY_CLIENT_ENTRY = "ctrl_velocity"
            self._DEFAULT_VOLTS_CLIENT_ENTRY = "ctrl_volts"

            self._MODULE_FILE_NAME = "step_direction.json"

        elif firmware.lower() == "servo":
            self._DEFAULT_CONTROL_CLIENT = "multi_turn_angle_control"
            self._DEFAULT_VELOCITY_CLIENT_ENTRY = "ctrl_velocity"
            self._DEFAULT_VOLTS_CLIENT_ENTRY = "ctrl_volts"

            self._MODULE_FILE_NAME = "servo.json"

        elif firmware.lower() == "pulsing":
            self._DEFAULT_CONTROL_CLIENT = "propeller_motor_control"
            self._DEFAULT_VELOCITY_CLIENT_ENTRY = "ctrl_velocity"
            self._DEFAULT_VOLTS_CLIENT_ENTRY = "ctrl_volts"

            self._MODULE_FILE_NAME = "pulsing.json"

        else:
            raise IqModuleError("'" + str(firmware) + "' firmware is not supported")

        # Pass the Super
        super().__init__(com, module_idn, clients_path)



import json
import os


class IqModuleJsonParser:
    @staticmethod
    def parse_default_modules(module_file_name: str):

        file_path = os.path.join(
            os.path.dirname(__file__), ("module_files/" + module_file_name)
        )

        with open(file_path) as json_file:
            module_file = json.load(json_file)

        return module_file

    @staticmethod
    def parse_module(module_file_path: str):

        with open(module_file_path) as json_file:
            module_file = json.load(json_file)

        return module_file



import os

from iqmotion.iq_devices.iq_module import IqModule
from iqmotion.communication.communicator import Communicator


# There are 3 options when it comes to adding extra clients
#
#       -Quick and dirty-
#   1.  clients_path {str: directory path}:   
#           Points to a directory containing client entry jsons
#           Note: this option will add every client entry in the folder
# 
#       -More Work but Flexible-
#   2.  extra_clients {list: [dir path, dir path, ...]}:
#           Contains a list of paths to each cleint entry you want to include
#           Note: you need to pass in an absolute paths
#
#       -Most Work but cleanest-
#   3.  custom_module_location {str: directory path}:
#           Commonly passed in as a cwd path, this can contain a module file
#           in addition to extra client files. 
#           Note: You need to have a folder named extra_client_files hosting your 
#                 client files
#           Example of 3rd Option
#           {
#             "clients": [
#               "brushless_drive",
#               "propeller_motor_control",
#               "anticogging",
#               "buzzer_control",
#               "esc_propeller_input_parser",
#               "hobby_input",
#               "persistent_memory",
#               "power_monitor",
#               "serial_interface",
#               "servo_input_parser",
#               "step_direction_input",
#               "system_control",
#               "temperature_estimator",
#               "temperature_monitor_uc"
#             ],
#             "extra_clients": [
#               "extra_client"
#             ]
#           }


class CustomIqModule(IqModule):
    _MODULE_FILE_NAME = ""

    def __init__(
        self,
        custom_module_location: str,
        com: Communicator,
        module_idn=0,
        clients_path: str=None,
        extra_clients: list=None,
    ):
        module_file_path = os.path.join(
            os.path.dirname(custom_module_location),
            ("module_files/" + self._MODULE_FILE_NAME),
        )

        super().__init__(com, 
                        module_idn, 
                        clients_path=clients_path,
                        extra_clients=extra_clients, 
                        module_file_path=module_file_path)

        # This will pull clients from the Module JSON 
        #       --Refer to example Below--
        if "extra_clients" in self._module_file_dict:
            for extra_client_name in self._module_file_dict["extra_clients"]:
                extra_client_file_path = os.path.join(
                    os.path.dirname(custom_module_location),
                    (f"extra_client_files/{extra_client_name}.json"),
                )
                self.add_client(extra_client_file_path)


